@@ -16,6 +16,9 @@ class StatStruct:
     is_symlink: bool
 
 
+default_db_path = os.path.join(jupyter_data_dir(), "file_id_manager.db")
+
+
 def log(log_before, log_after):
     """Decorator that accepts two functions which build a log string to be
     logged to INFO before and after the target method executes. The functions
@@ -35,7 +38,7 @@ def log(log_before, log_after):
 
 class FileIdManager(LoggingConfigurable):
     """
-    Manager that supports tracks files across their lifetime by associating
+    Manager that supports tracking files across their lifetime by associating
     each with a unique file ID, which is maintained across filesystem operations.
 
     Notes
@@ -54,7 +57,7 @@ class FileIdManager(LoggingConfigurable):
     )
 
     db_path = Unicode(
-        default_value=os.path.join(jupyter_data_dir(), "file_id_manager.db"),
+        default_value=default_db_path,
         help=(
             "The path of the DB file used by `FileIdManager`. "
             "Defaults to `jupyter_data_dir()/file_id_manager.db`."
@@ -62,16 +65,18 @@ class FileIdManager(LoggingConfigurable):
         config=True,
     )
 
-    def __init__(self, **kwargs):
+    def __init__(self, *args, **kwargs):
         # pass args and kwargs to parent Configurable
-        super().__init__(**kwargs)
+        super().__init__(*args, **kwargs)
         # initialize instance attrs
         self._update_cursor = False
         # initialize connection with db
         self.con = sqlite3.connect(self.db_path)
-        self.log.info("FileIdManager : Configured root dir: %s" % self.root_dir)
-        self.log.info("FileIdManager : Configured database path: %s" % self.db_path)
+        self.log.info(f"FileIdManager : Configured root dir: {self.root_dir}")
+        self.log.info(f"FileIdManager : Configured database path: {self.db_path}")
         self.log.info("FileIdManager : Creating File ID tables and indices")
+        # do not allow reads to block writes. required when using multiple processes
+        self.con.execute("PRAGMA journal_mode = WAL")
         self.con.execute(
             "CREATE TABLE IF NOT EXISTS Files("
             "id INTEGER PRIMARY KEY AUTOINCREMENT, "
@@ -93,9 +98,9 @@ class FileIdManager(LoggingConfigurable):
     @validate("root_dir", "db_path")
     def _validate_abspath_traits(self, proposal):
         if proposal["value"] is None:
-            raise TraitError("FileIdManager : %s must not be None" % proposal["trait"].name)
+            raise TraitError(f"FileIdManager : {proposal['trait'].name} must not be None")
         if not os.path.isabs(proposal["value"]):
-            raise TraitError("FileIdManager : %s must be an absolute path" % proposal["trait"].name)
+            raise TraitError(f"FileIdManager : {proposal['trait'].name} must be an absolute path")
         return self._normalize_path(proposal["value"])
 
     def _index_all(self):
@@ -110,16 +115,13 @@ class FileIdManager(LoggingConfigurable):
             for entry in scan_iter:
                 if entry.is_dir():
                     self._index_dir_recursively(entry.path, self._stat(entry.path))
-        scan_iter.close()
 
-    def _sync_all(self, id=None):
+    def _sync_all(self):
         """
         Syncs Files table with the filesystem and ensures that the correct path
         is associated with each file ID. Does so by iterating through all
         indexed directories and syncing the contents of all dirty directories.
 
-        Parameters
-        ----------
         Notes
         -----
         A dirty directory is a directory that is either:
@@ -129,7 +131,7 @@ class FileIdManager(LoggingConfigurable):
         Dirty directories contain possibly indexed but moved files as children.
         Hence we need to call _sync_file() on their contents via _sync_dir().
         Indexed directories with mtime difference are handled in this method
-        body.  Unindexed dirty directories are handled immediately when
+        body. Unindexed dirty directories are handled immediately when
         encountered in _sync_dir().
 
         If a directory was indexed-but-moved, the existing cursor may contain
@@ -137,12 +139,12 @@ class FileIdManager(LoggingConfigurable):
         _sync_file(). Hence the cursor needs to be redefined if
         self._update_cursor is set to True by _sync_file().
         """
-        cursor = self.con.execute("SELECT id, path, mtime FROM Files WHERE is_dir = 1")
+        cursor = self.con.execute("SELECT path, mtime FROM Files WHERE is_dir = 1")
         self._update_cursor = False
         dir = cursor.fetchone()
 
         while dir:
-            id, path, old_mtime = dir
+            path, old_mtime = dir
             stat_info = self._stat(path)
 
             # ignores directories that no longer exist
@@ -155,12 +157,15 @@ class FileIdManager(LoggingConfigurable):
 
             if dir_dirty:
                 self._sync_dir(path)
-                self._update(id, stat_info)
+                # prefer index over _sync_file() as it ensures directory is
+                # stored back into the Files table in the case of `mtime`
+                # mismatch, which results in deleting the old record.
+                self.index(path, stat_info, commit=False)
 
             # check if cursor should be updated
             if self._update_cursor:
                 self._update_cursor = False
-                cursor = self.con.execute("SELECT id, path, mtime FROM Files WHERE is_dir = 1")
+                cursor = self.con.execute("SELECT path, mtime FROM Files WHERE is_dir = 1")
 
             dir = cursor.fetchone()
 
@@ -189,6 +194,23 @@ class FileIdManager(LoggingConfigurable):
 
         scan_iter.close()
 
+    def _check_timestamps(self, stat_info):
+        """Returns True if the timestamps of a file match those recorded in the
+        Files table. Returns False otherwise."""
+
+        src = self.con.execute(
+            "SELECT crtime, mtime FROM Files WHERE ino = ?", (stat_info.ino,)
+        ).fetchone()
+
+        # if no record with matching ino, then return None
+        if not src:
+            return False
+
+        src_crtime, src_mtime = src
+        src_timestamp = src_crtime if src_crtime is not None else src_mtime
+        dst_timestamp = stat_info.crtime if stat_info.crtime is not None else stat_info.mtime
+        return src_timestamp == dst_timestamp
+
     def _sync_file(self, path, stat_info):
         """
         Syncs the file at `path` with the Files table by detecting whether the
@@ -222,34 +244,27 @@ class FileIdManager(LoggingConfigurable):
             return None
 
         src = self.con.execute(
-            "SELECT id, path, crtime, mtime FROM Files WHERE ino = ?", (stat_info.ino,)
+            "SELECT id, path FROM Files WHERE ino = ?", (stat_info.ino,)
         ).fetchone()
 
-        # if no record with matching ino, then return None
-        if not src:
+        # if ino is not in database, return None
+        if src is None:
+            return None
+        id, old_path = src
+
+        # if timestamps don't match, delete existing record and return None
+        if not self._check_timestamps(stat_info):
+            self.con.execute("DELETE FROM Files WHERE id = ?", (id,))
             return None
 
-        id, old_path, src_crtime, src_mtime = src
-        src_timestamp = src_crtime if src_crtime is not None else src_mtime
-        dst_timestamp = stat_info.crtime if stat_info.crtime is not None else stat_info.mtime
+        # otherwise update existing record with new path, moving any indexed
+        # children if necessary. then return its id
+        self._update(id, path=path)
+        if stat_info.is_dir and old_path != path:
+            self._move_recursive(old_path, path)
+            self._update_cursor = True
 
-        # if record has identical ino and crtime/mtime to an existing record,
-        # update it with new path, returning its id
-        if src_timestamp == dst_timestamp:
-            self._update(id, path=path)
-
-            # update paths of indexed children under moved directories and set
-            # self._update_cursor to True
-            if stat_info.is_dir and old_path != path:
-                self._move_recursive(old_path, path)
-                self._update_cursor = True
-
-            return id
-
-        # otherwise delete the existing record with identical `ino`, since inos
-        # must be unique. then return None
-        self.con.execute("DELETE FROM Files WHERE id = ?", (id,))
-        return None
+        return id
 
     def _normalize_path(self, path):
         """Normalizes a given file path."""
@@ -280,7 +295,7 @@ class FileIdManager(LoggingConfigurable):
         return stat_info
 
     def _stat(self, path):
-        """Returns stat info on a path in a StatStruct object.Returns None if
+        """Returns stat info on a path in a StatStruct object. Returns None if
         file does not exist at path."""
         try:
             raw_stat = os.lstat(path)
@@ -291,7 +306,14 @@ class FileIdManager(LoggingConfigurable):
 
     def _create(self, path, stat_info):
         """Creates a record given its path and stat info. Returns the new file
-        ID."""
+        ID.
+
+        Notes
+        -----
+        - Because of the uniqueness constraint on `ino`, this method is
+        dangerous and may throw a runtime error if the file is not guaranteed to
+        have a unique `ino`.
+        """
         cursor = self.con.execute(
             "INSERT INTO Files (path, ino, crtime, mtime, is_dir) VALUES (?, ?, ?, ?, ?)",
             (path, stat_info.ino, stat_info.crtime, stat_info.mtime, stat_info.is_dir),
@@ -300,12 +322,20 @@ class FileIdManager(LoggingConfigurable):
         return cursor.lastrowid
 
     def _update(self, id, stat_info=None, path=None):
-        """Updates a record given its file ID and stat info."""
-        # updating `ino` and `crtime` is a conscious design decision because
-        # this method is called by `move()`. these values are only preserved by
-        # fs moves done via the `rename()` syscall, like `mv`. we don't care how
-        # the contents manager moves a file; it could be deleting and creating a
-        # new file (which will change the stat info).
+        """Updates a record given its file ID and stat info.
+
+        Notes
+        -----
+        - Updating `ino` and `crtime` is a conscious design decision because
+        this method is called by `move()`. These values are only preserved by
+        fs moves done via the `rename()` syscall, like `mv`. We don't care how
+        the contents manager moves a file; it could be deleting and creating a
+        new file (which will change the stat info).
+
+        - Because of the uniqueness constraint on `ino`, this method is
+        dangerous and may throw a runtime error if the file is not guaranteed to
+        have a unique `ino`.
+        """
         if stat_info and path:
             self.con.execute(
                 "UPDATE Files SET ino = ?, crtime = ?, mtime = ?, path = ? WHERE id = ?",
@@ -327,6 +357,19 @@ class FileIdManager(LoggingConfigurable):
             )
             return
 
+    def sync_all(self):
+        """
+        Syncs Files table with the filesystem and ensures that the correct path
+        is associated with each file ID. Required to be called manually
+        beforehand if `sync=False` is passed to `get_path()`.
+
+        Notes
+        -----
+        - Identical to `_sync_all()`, but is public and commits the transaction
+        upon invocation. See `_sync_all()` for more implementation details."""
+        self._sync_all()
+        self.con.commit()
+
     def index(self, path, stat_info=None, commit=True):
         """Returns the file ID for the file at `path`, creating a new file ID if
         one does not exist. Returns None only if file does not exist at path."""
@@ -364,19 +407,49 @@ class FileIdManager(LoggingConfigurable):
         self.con.commit()
         return id
 
-    def get_path(self, id):
-        """Retrieves the file path associated with a file ID. Returns None if
-        the ID does not exist in the Files table or if the corresponding path no
-        longer has a file."""
-        self._sync_all()
-        self.con.commit()
-        row = self.con.execute("SELECT path FROM Files WHERE id = ?", (id,)).fetchone()
-        path = row and row[0]
+    def get_path(self, id, sync=True):
+        """Retrieves the file path associated with a file ID. The file path is
+        relative to `self.root_dir`. Returns None if the ID does not
+        exist in the Files table, if the path no longer has a
+        file, or if the path is not a child of `self.root_dir`.
 
-        # if no record associated with ID or if file no longer exists at path, return None
-        if path is None or self._stat(path) is None:
+        Parameters
+        ----------
+        sync : bool
+            Whether to invoke `sync_all()` automatically prior to ID lookup.
+
+        Notes
+        -----
+        - For performance reasons, if this method must be called multiple times
+        in serial, then it is best to first invoke `sync_all()` and then call
+        this method with the argument `sync=False`.
+        """
+        if sync:
+            self.sync_all()
+
+        row = self.con.execute("SELECT path, ino FROM Files WHERE id = ?", (id,)).fetchone()
+
+        # if no record associated with ID, return None
+        if row is None:
+            return None
+        path, ino = row
+
+        # if file no longer exists at path, return None
+        stat_info = self._stat(path)
+        if stat_info is None:
             return None
 
+        # if inode numbers or timestamps of the file and record don't match,
+        # return None
+        if ino != stat_info.ino or not self._check_timestamps(stat_info):
+            return None
+
+        # if path is not relative to `self.root_dir`, return None.
+        if os.path.commonpath([self.root_dir, path]) != self.root_dir:
+            return None
+
+        # finally, convert the path to a relative one.
+        path = os.path.relpath(path, self.root_dir)
         return path
 
     def _move_recursive(self, old_path, new_path):
@@ -484,6 +557,34 @@ class FileIdManager(LoggingConfigurable):
         self.con.execute("DELETE FROM Files WHERE path = ?", (path,))
         self.con.commit()
 
+    def save(self, path):
+        """Handles file saves (edits) by updating recorded stat info.
+
+        Notes
+        -----
+        - This assumes that the file was present prior to the save event. That
+        means it's technically possible to fool this method by deleting and
+        creating a new file at the same path out-of-band, and then update it via
+        JupyterLab.  This would (wrongly) preserve the association b/w the old
+        file ID and the current path rather than create a new file ID.
+        """
+        path = self._normalize_path(path)
+
+        # look up record by ino and path
+        stat_info = self._stat(path)
+        row = self.con.execute(
+            "SELECT id FROM Files WHERE ino = ? AND path = ?", (stat_info.ino, path)
+        ).fetchone()
+
+        # if no record exists, return early
+        if row is None:
+            return
+
+        # otherwise, update the stat info
+        (id,) = row
+        self._update(id, stat_info)
+        self.con.commit()
+
     def __del__(self):
         """Cleans up `FileIdManager` by committing any pending transactions and
         closing the connection."""
